/**
 * LDT
 *
 * written by Colin Kuebler 2012
 * refactored by Deema Ywanov 2015
 */

const on = require("emmy/on");
const getCaret = require("caret-position2/get");
const setCaret = require("caret-position2/set");

/** @class TextareaDecorator */
class TextareaDecorator {
  constructor(textarea, parser) {
    if (!(this instanceof TextareaDecorator)) {
      return new TextareaDecorator(textarea, parser);
    }

    var self = this;

    //create parser
    if (!(parser instanceof Parser)) {
      parser = new Parser(parser);
    }
    self.parser = parser;

    // construct DOM
    self.highlighter = document.createElement("pre");
    self.highlighter.classList.add("ldt");
    self.highlighter.contentEditable = true;
    self.highlighter.tabindex = 0;

    //keep textarea as a stub
    self.textarea = textarea;
    self.textarea.classList.add("ldt-textarea");
    self.textarea.spellcheck = false;
    self.textarea.parentNode.insertBefore(self.highlighter, self.textarea);

    // detect all changes to the textarea,
    // including keyboard input, cut/copy/paste, drag & drop, etc
    on(self.textarea, "input", function() {
      self.update();
    });
    on(self.textarea, "scroll", function() {
      self.highlighter.scrollTop = self.textarea.scrollTop;
      self.highlighter.scrollLeft = self.textarea.scrollLeft;
      self.update();
    });
    //update on resize
    on(self.textarea, "mousemove touchmove", function() {
      self.update();
    });

    // initial highlighting
    self.update();

    //prevent tab switching
    self.textarea.addEventListener("keydown", function(e) {
      if (e.which !== 9) {
        return;
      }

      e.preventDefault();

      var caret = getCaret(self.textarea);
      var value = self.textarea.value;
      var selectedValue = value.slice(caret.start, caret.end);

      //if selected value includes line breaks - do line tabbing
      var lines = selectedValue.split(/\r\n|\r|\n/).length;
      if (e.shiftKey || lines > 1) {
        //expand selection to the first newline
        var before = value.slice(0, caret.start);
        var firstBreakIdx = before.lastIndexOf("\n");
        if (firstBreakIdx === -1) {
          firstBreakIdx = 0;
        }
        selectedValue = value.slice(firstBreakIdx, caret.end);

        //if shift is pressed - untab each selected row
        if (e.shiftKey) {
          var tabLines = selectedValue.split(/\n\t/).length - 1;
          var modFirst = /^\n\t/.test(selectedValue);
          selectedValue = selectedValue.replace(/\n\t/g, "\n");
          self.textarea.value =
            value.slice(0, firstBreakIdx) +
            selectedValue +
            value.slice(caret.end);
          setCaret(
            self.textarea,
            caret.start - (modFirst ? 1 : 0),
            caret.end - tabLines
          );
        }
        //tab each row
        else {
          selectedValue = selectedValue.replace(/\n/g, "\n\t");
          self.textarea.value =
            value.slice(0, firstBreakIdx) +
            selectedValue +
            value.slice(caret.end);
          setCaret(self.textarea, caret.start + 1, caret.end + lines);
        }
      }
      //just insert simple tab
      else {
        self.textarea.value =
          value.slice(0, caret.start) +
          "\t" +
          selectedValue +
          value.slice(caret.end);
        setCaret(self.textarea, caret.start + 1, caret.end + 1);
      }

      self.update();
    });
  }

  /**
   * Mimic textarea style
   * Update syntax
   */
  update() {
    var self = this;

    if (self.plannedUpdate) {
      return;
    }

    self.plannedUpdate = true;

    requestAnimationFrame(function() {
      self.plannedUpdate = false;

      var style = getComputedStyle(self.textarea);

      //copy textarea style
      self.highlighter.style.fontSize = style.fontSize;
      self.highlighter.style.borderWidth = style.borderWidth;
      self.highlighter.style.height = style.height;
      self.highlighter.style.width = style.width;
      self.highlighter.style.margin = style.margin;
      self.highlighter.style.padding = style.padding;
      self.highlighter.style.wordWrap = style.wordWrap;
      self.highlighter.style.wordBreak = style.wordBreak;
      self.highlighter.style.whiteSpace = style.whiteSpace;
      self.highlighter.style.wordSpacing = style.wordSpacing;
      self.highlighter.style.top = self.textarea.offsetTop + "px";
      self.highlighter.style.left = self.textarea.offsetLeft + "px";

      //keep textarea updated
      var input = self.textarea.value;

      //coloring algorithm
      var oldTokens = self.highlighter.childNodes;
      var newTokens = self.parser.tokenize(input);
      var firstDiff, lastDiffNew, lastDiffOld;
      // find the first difference
      for (
        firstDiff = 0;
        firstDiff < newTokens.length && firstDiff < oldTokens.length;
        firstDiff++
      )
        if (newTokens[firstDiff] !== oldTokens[firstDiff].textContent) break;
      // trim the length of output nodes to the size of the input
      while (newTokens.length < oldTokens.length)
        self.highlighter.removeChild(oldTokens[firstDiff]);
      // find the last difference
      for (
        lastDiffNew = newTokens.length - 1, lastDiffOld = oldTokens.length - 1;
        firstDiff < lastDiffOld;
        lastDiffNew--, lastDiffOld--
      )
        if (newTokens[lastDiffNew] !== oldTokens[lastDiffOld].textContent)
          break;
      // update modified spans
      for (; firstDiff <= lastDiffOld; firstDiff++) {
        oldTokens[firstDiff].className = self.parser.identify(
          newTokens[firstDiff]
        );
        oldTokens[firstDiff].textContent = oldTokens[firstDiff].innerText =
          newTokens[firstDiff];
      }
      // add in modified spans
      for (
        var insertionPt = oldTokens[firstDiff] || null;
        firstDiff <= lastDiffNew;
        firstDiff++
      ) {
        var span = document.createElement("span");
        span.className = self.parser.identify(newTokens[firstDiff]);
        span.textContent = span.innerText = newTokens[firstDiff];
        self.highlighter.insertBefore(span, insertionPt);
      }
    });
  }
}

/**
 * Generates a tokenizer from regular expressions for TextareaDecorator
 */
class Parser {
  constructor(rules, i) {
    var self = this;

    // variables used internally
    self.i = i ? "i" : "";

    self.parseRE = null;
    self.ruleSrc = [];
    self.ruleMap = {};

    self.add(rules);

    return self;
  }

  add(rules) {
    var self = this;

    for (var rule in rules) {
      var s = rules[rule].source;
      self.ruleSrc.push(s);
      self.ruleMap[rule] = new RegExp("^(" + s + ")$", self.i);
    }

    self.parseRE = new RegExp(self.ruleSrc.join("|"), "g" + self.i);

    return self;
  }

  tokenize(input) {
    var self = this;

    return input.match(self.parseRE) || [];
  }

  identify(token) {
    var self = this;

    for (var rule in self.ruleMap) {
      if (self.ruleMap[rule].test(token)) {
        return rule;
      }
    }
  }
}

module.exports = TextareaDecorator
